// Copyright (C) 2019-2021  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package zdata

//go:generate ../set/gen-set zdata ZBigFile *ZBigFile zset_bigfile.go

import (
	"context"
	"fmt"
	"runtime"
	"sync"

	"lab.nexedi.com/kirr/go123/xerr"
	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"

	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/set"
	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xbtree"
	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xtail"
	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xzodb"
)

type SetI64 = set.SetI64

// ΔFtail represents tail of revisional changes to files.
//
// It semantically consists of
//
//	[]δF			; rev ∈ (tail, head]
//
// where δF represents a change in files space
//
//	δF:
//		.rev↑
//		{} file -> {}blk
//
// Only files and blocks explicitly requested to be tracked are guaranteed to
// be present. In particular a block that was not explicitly requested to be
// tracked, even if it was changed in δZ, is not guaranteed to be present in δF.
//
// ΔFtail provides the following operations:
//
//   .Track(file, blk, path, zblk)	- add file and block reached via BTree path to tracked set.
//
//   .Update(δZ) -> δF				- update files δ tail given raw ZODB changes
//   .ForgetPast(revCut)			- forget changes past revCut
//   .SliceByRev(lo, hi) -> []δF		- query for all files changes with rev ∈ (lo, hi]
//   .SliceByFileRev(file, lo, hi) -> []δfile	- query for changes of file with rev ∈ (lo, hi]
//   .LastBlkRev(file, #blk, at)		- query for what is last revision that changed
//						  file[#blk] as of @at database state.
//
// where δfile represents a change to one file
//
//	δfile:
//		.rev↑
//		{}blk
//
// XXX concurrent use
//
// See also zodb.ΔTail
type ΔFtail struct {
	// ΔFtail merges ΔBtail with history of ZBlk
	δBtail  *xbtree.ΔBtail
	fileIdx map[zodb.Oid]SetZBigFile // tree-root -> {} ZBigFile as of @head

	// XXX don't need vδF - everything is reconstructed at runtime from .δBtail.vδT
	//     this way we also don't need to keep up updating vδF from vδT on its rebuild during.
	// data with δF changes. Actual for part of tracked set that was taken
	// into account.
	vδF []ΔF

	// tracked ZBlk that are not yet taken into account in current vδF.
	// grows on new track requests; flushes on queries and update.
	trackNew map[*ZBigFile]map[zodb.Oid]*zblkInΔFtail // {} file -> {} oid -> zblk
}

// ΔF represents a change in files space.
type ΔF struct {
	Rev    zodb.Tid
	ByFile map[*ZBigFile]*ΔFile // file -> δfile
}

// ΔFile represents a change to one file.
type ΔFile struct {
	Rev    zodb.Tid
	Blocks SetI64 // changed blocks
	Size   bool   // whether file size changed
}

// zblkInΔFtail is ΔFtail-related volatile data embedded into ZBlk*.
//
// The data is preserved even when ZBlk comes to ghost state, but is lost if
// ZBlk is garbage collected. The data is transient - it is _not_ included into
// persistent state.
type zblkInΔFtail struct {
	mu sync.Mutex // to support multiple concurrent loaders

	// XXX change vvv to intree_parent? {} Bucket -> set(#blk)
	// (this is uniform with building in-RAM reverse child->parents relation for
	//  tree nodes and for tree_root->file)

	// with which files/blocks this ZBlk is associated with as of @head state
	infile map[*ZBigFile]SetI64 // {} file -> set(#blk)
}

type _ZBlkInΔFtail interface { inΔFtail() *zblkInΔFtail }
func (z *zblkInΔFtail) inΔFtail() *zblkInΔFtail { return z }

// NewΔFtail creates new ΔFtail object.
//
// Initial tracked set is empty.
// Initial coverage of created ΔFtail is (at₀, at₀].
//
// db will be used by ΔFtail to open database connections to load data from
// ZODB when needed.
func NewΔFtail(at0 zodb.Tid, db *zodb.DB) *ΔFtail {
	return &ΔFtail{
		δBtail:   xbtree.NewΔBtail(at0, db),
		fileIdx:  make(map[zodb.Oid]SetZBigFile),
		trackNew: make(map[*ZBigFile]map[zodb.Oid]*zblkInΔFtail),
	}
}

// (tail, head] coverage
func (δFtail *ΔFtail) Head() zodb.Tid { return δFtail.δBtail.Head() }
func (δFtail *ΔFtail) Tail() zodb.Tid { return δFtail.δBtail.Tail() }


// Track associates file[blk]@head with tree path and zblk object.
//
// zblk can be nil, which represents a hole.
// XXX blk=-1 is used for tracking after Size (no zblk is accessed at all).
//
// XXX Track adds tree path to tracked set and associates path root with file.
//
// XXX text
//
// XXX objects in path and zblk must be with .PJar().At() == .head
//
// A root can be associated with several files (each provided on different Track call).
func (δFtail *ΔFtail) Track(file *ZBigFile, blk int64, path []btree.LONode, zblk ZBlk) {
	if blk == -1 {
		// XXX blk = ∞ from beginning ?
		blk = xbtree.KeyMax
	}
	err := δFtail.δBtail.Track(blk, zblk != nil, path)
	if err != nil {
		panic(err) // XXX -> error? errctx
	}
	root := path[0].(*btree.LOBTree)
	files, ok := δFtail.fileIdx[root.POid()]
	if !ok {
		files = SetZBigFile{}
		δFtail.fileIdx[root.POid()] = files
	}
	files.Add(file)

	// associate zblk with file, if it was not hole
	if zblk != nil {
		z := zblk.inΔFtail()
		z.mu.Lock()
		blocks, ok := z.infile[file]
		if !ok {
			blocks = make(SetI64, 1)
			if z.infile == nil {
				z.infile = make(map[*ZBigFile]SetI64)
			}
			z.infile[file] = blocks
		}
		blocks.Add(blk)
		z.mu.Unlock()

		// XXX locking
		if !ok {
			// zblk was not associated with this file
			zt := δFtail.trackNew[file]
			if zt == nil {
				zt = make(map[zodb.Oid]*zblkInΔFtail, 1)
				δFtail.trackNew[file] = zt
			}
			zt[zblk.POid()] = z
		}
	}

	// XXX mark something dirty so that LastBlkRev and Slice* know what to rebuild?

	// XXX debug
/*
	leaf := path[len(path)-1].(*btree.LOBucket)
	for _, e := range leaf.Entryv() { // XXX activate
		δFtail.tracked.Add(e.Key())
	}
*/
}

// Update updates δFtail given raw ZODB changes.
//
// It returns change in files space that corresponds to δZ.
//
// δZ should include all objects changed by ZODB transaction.
//
// Zhead must be active connection at δFtail.Head() database state.
// Objects in Zhead must not be modified.
// During call to Update zhead must not be otherwise used - even for reading.
func (δFtail *ΔFtail) Update(δZ *zodb.EventCommit, zhead *xzodb.ZConn) (_ ΔF, err error) {
	defer xerr.Contextf(&err, "ΔFtail update %s -> %s", δFtail.Head(), δZ.Tid)
	// XXX δFtail.update() first?

	// XXX verify zhead.At() == δFtail.Head()

	δB, err := δFtail.δBtail.Update(δZ)
	if err != nil {
		return ΔF{}, err
	}

	δF := ΔF{Rev: δB.Rev, ByFile: make(map[*ZBigFile]*ΔFile)}

	// take btree changes into account
	for root, δt := range δB.ByRoot {
		files := δFtail.fileIdx[root]
		if len(files) == 0 {
			panicf("BUG: ΔFtail: root<%s> -> ø files", root)
		}
		for file := range files {
			δfile, ok := δF.ByFile[file]
			if !ok {
				δfile = &ΔFile{Rev: δF.Rev, Blocks: make(SetI64)}
				δF.ByFile[file] = δfile
			}
			for blk /*, zblk*/ := range δt {
				// FIXME stub - need to take both keys and zblk changes into account
				// XXX document, and in particular how to include atTail
				δfile.Blocks.Add(blk)
			}

			// TODO invalidate .size only if key >= maxkey was changed (size increase),
			// or if on the other hand maxkey was deleted (size decrese).
			//
			// XXX currently we invalidate size on any topology change.
			δfile.Size = true
		}
	}

	// take zblk changes into account
	for _, oid := range δZ.Changev {
		// XXX cache lock/unlock
		obj := zhead.Cache().Get(oid)
		if obj == nil {
			//fmt.Printf("%s: not in cache\n", oid)
			continue // nothing to do - see invariant
		}

		//fmt.Printf("%s:     in cache (%s)\n", oid, typeOf(obj))

		switch obj := obj.(type) {
		case ZBlk:	// ZBlk*
			// z.infile locking: since we write-locked head.zheadMu
			// - no other fuse reads are running, and thus no one
			// is mutating z.infile. XXX recheck
			z := obj.inΔFtail()
			for file, blocks := range z.infile {
				δfile, ok := δF.ByFile[file]
				if !ok {
					δfile = &ΔFile{Rev: δF.Rev, Blocks: make(SetI64)}
					δF.ByFile[file] = δfile
				}

				δfile.Blocks.Update(blocks)
			}

			// XXX update z.infile according to btree changes

		case *ZBigFile:
			// TODO check that .blksize and .blktab (it is only
			// persistent reference) do not change.

			return ΔF{}, fmt.Errorf("ZBigFile<%s> changed @%s", oid, δZ.Tid)
		}

		// make sure obj won't be garbage-collected until we finish handling it.
		runtime.KeepAlive(obj)
	}

	δFtail.vδF = append(δFtail.vδF, δF)
	return δF, nil
}

// XXX kill after vδF is gone
// update processes new track requests and updates vδF.
//
// If file != nil only track requests related to file are processed.
// Otherwise all track requests are processed.
func (δFtail *ΔFtail) update(file *ZBigFile) {
	if file == nil {
		panic("TODO")
	}

	// let's see if we need to rebuild .vδF due to not-yet processed track requests

	// XXX locking
	// XXX dumb
	zt, dirty := δFtail.trackNew[file]
	if !dirty {
		return
	}

	delete(δFtail.trackNew, file)
	// XXX unlock here

	for i, δZ := range δFtail.δBtail.ΔZtail().Data() {
		δF := δFtail.vδF[i]
		// XXX assert δF.Rev == δZ.Rev

		for _, oid := range δZ.Changev {
			z, ok := zt[oid]
			if !ok {
				continue
			}

			// XXX locking
			// XXX -> func δF.δfile(file) ?
			δfile, ok := δF.ByFile[file]
			if !ok {
				δfile = &ΔFile{Rev: δF.Rev, Blocks: make(SetI64)}
				δF.ByFile[file] = δfile
			}

			δfile.Blocks.Update(z.infile[file])
		}
	}
}

// ForgetPast discards all δFtail entries with rev ≤ revCut.
func (δFtail *ΔFtail) ForgetPast(revCut zodb.Tid) {
	δFtail.δBtail.ForgetPast(revCut)
}

// XXX don't need
func (δFtail *ΔFtail) SliceByRev(lo, hi zodb.Tid) /*readonly*/ []ΔF {
	xtail.AssertSlice(δFtail, lo, hi)
	panic("TODO")
}

// SliceByFileRev returns history of file changes in (lo, hi] range.
//
// it must be called with the following condition:
//
//	tail ≤ lo ≤ hi ≤ head
//
// the caller must not modify returned slice.
//
// Note: contrary to regular go slicing, low is exclusive while high is inclusive.
func (δFtail *ΔFtail) SliceByFileRev(file *ZBigFile, lo, hi zodb.Tid) /*readonly*/[]*ΔFile {
	xtail.AssertSlice(δFtail, lo, hi)

	// FIXME rework to just query .δBtail.SliceByRootRev(file.blktab, lo, hi) +
	//       merge δZBlk history with that.

	// XXX locking?
	δFtail.update(file)

	// find vδF range corresponding to (lo, hi]
	// XXX linear scan
	vδF := δFtail.vδF
	if len(vδF) == 0 {
		return nil
	}

	// find max j : [j].rev ≤ hi		XXX linear scan -> binary search
	j := len(vδF)-1
	for ; j >= 0 && vδF[j].Rev > hi; j-- {}
	if j < 0 {
		return nil // ø
	}

	// find max i : [i].rev > low		XXX linear scan -> binary search
	i := j
	for ; i >= 0 && vδF[i].Rev > lo; i-- {}
	i++

	vδF = vδF[i:j+1]

	// filter found changed to have only file-related bits
	var vδfile []*ΔFile
	for _, δF := range vδF {
		δfile, ok := δF.ByFile[file]
		if ok {
			vδfile = append(vδfile, δfile)
		}
	}

	// XXX merge into vδF zblk from not yet handled tracked part

	return vδfile


	// merging tree (δT) and Zblk (δZblk) histories into file history (δFile):

	// δT		────────·──────────────·─────────────────·────────────
	//		        │              │
	//		        ↓              │
	// δZblk₁	────────────────o───────────────────o─────────────────
	//                                     |
	//                                     ↓
	// δZblk₂	────────────x────────────────x────────────────────────
	//
	//
	// δFile	────────o───────o──────x─────x────────────────────────

/*
	vδZ := δFtail.δBtail.ΔZtail().SliceByRev(lo, hi)

	// XXX stub that takes only ZBlk changes into account
	// XXX dumb
	for _, δZ := range vδZ {
	}
*/


/*
	// XXX activate zfile?
	vδT := δFtail.δBtail.SliceByRootRev(file.zfile.blktab, lo, hi)

	// state of `{} blk -> zblk` as we are scanning ↓
	δblktab := map[int64]struct {
		zblk   zodb.Oid // blk points to this zblk
		lo, hi zodb.Tid // blk points to zblk during [lo, hi)
	}{}

	iz := len(vδZ) - 1
	it := len(vδT) - 1
	for (iz >= 0 && it >= 0) {	// XXX -> ||
		δZ := vδZ[iz]
		δT := vδT[it]
		if δZ.Rev >= δT.Rev {
			for _, oid := range δZ.Changev {
				// XXX oid -> tracked ZBlk?
				// ZBlk -> bound to {}blk @head
				for blk := range boundToAtHead {
					if !δblktab.Has(blk) {
						δblktab[blk] = oid
					}
				}
			}
		}
		if δT.Rev >= δZ.Rev {
			...
		}
	}
*/

}

// XXX rename -> BlkRevAt
// LastBlkRev returns last revision that changed file[blk] as of @at database state.
//
// if exact=False - what is returned is only an upper bound for last block revision.
//
// zf must be from @head
// at must ∈ (tail, head]
// blk must be tracked
//
// XXX +ctx, error	rebuild []δF here
func (δFtail *ΔFtail) LastBlkRev(ctx context.Context, zf *ZBigFile, blk int64, at zodb.Tid) (_ zodb.Tid, exact bool) {
	//defer xerr.Contextf(&err, "")	// XXX text

	// XXX assert δFtail == f.head.bfdir.δFtail ?

	// XXX tabRev -> treeRev ?
	// XXX activate zfile?
	zblkOid, ok, tabRev, tabRevExact, err := δFtail.δBtail.GetAt(ctx, zf.blktab, blk, at)
	if err != nil {
		panic(err) // XXX
	}

	// block was removed
	// XXX or not in tracked set?
	if !ok {
		return tabRev, tabRevExact
	}

	// blktab[blk] was changed to point to a zblk @rev.
	// blk revision is max rev and when zblk changed last in (rev, at] range.
	//
	// XXX need to use full δZ, not only connected to tracked subset?
	zblkRev, zblkRevExact := δFtail.δBtail.ΔZtail().LastRevOf(zblkOid, at)
	if zblkRev > tabRev {
		return zblkRev, zblkRevExact
	} else {
		return tabRev, tabRevExact
	}
}
